@@ -22,7 +22,22 @@ impl Parse for ContextType {
     }
 }
 
-#[proc_macro_derive(Server, attributes(context_type))]
+struct PaymentWebhookHandler {
+    name: Ident,
+}
+
+impl Parse for PaymentWebhookHandler {
+    fn parse(input: ParseStream) -> Result<Self> {
+        let content;
+        parenthesized!(content in input);
+
+        Ok(PaymentWebhookHandler {
+            name: content.parse()?,
+        })
+    }
+}
+
+#[proc_macro_derive(Server, attributes(context_type, payment_webhook_handler))]
 pub fn checkout_server_derive(input: TokenStream) -> TokenStream {
     let ast = parse(input).unwrap();
 
@@ -74,6 +89,72 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
     let common_lifetime_def = syn::LifetimeDef::new(syn::parse_quote!('common));
     let common_lifetime = common_lifetime_def.lifetime;
 
+    let maybe_pwh_attr = ast
+        .attrs
+        .iter()
+        .filter(|a| {
+            a.path.segments.len() == 1 && a.path.segments[0].ident == "payment_webhook_handler"
+        })
+        .nth(0);
+
+    let (payment_webhook_endpoint_def, payment_webhook_endpoint_reg) = match maybe_pwh_attr {
+        Some(pwh_attr) => {
+            let pwh: PaymentWebhookHandler =
+                parse2(pwh_attr.tokens.clone()).expect("invalid Payment Webhook Handler");
+            let pwh_name = pwh.name;
+
+            let endpoint_definition = quote! {
+                #[endpoint {
+                    method = POST,
+                    path = "/payment_webhook",
+                }]
+                pub async fn payment_webhook(
+                    req_ctx: Arc<RequestContext<ServerContext>>,
+                    body: TypedBody<serde_json::Value>,
+                ) -> Result<HttpResponseOk<()>, HttpError> {
+                    let internal_ctx = req_ctx
+                        .context()
+                        .internal_context
+                        .clone();
+                    let mut ctx = req_ctx
+                        .context()
+                        .context_provider
+                        .new_context()
+                        .await
+                        .unwrap();
+                    let req = req_ctx.request.clone();
+                    let guard = req.lock().await;
+                    let val = guard.deref();
+                    let headers = val.headers().clone();
+                    let data = body.into_inner();
+
+                    let (correlation_id, args) = super::#pwh_name(headers, data).await.map_err(|e| {
+                        let out: HttpError = e.into();
+                        out
+                    })?;
+                    ctx.start_transaction().await.unwrap();
+                    match callback_handler(&internal_ctx, &mut ctx, &correlation_id, args).await {
+                        Ok(_) => {
+                            ctx.commit_transaction().await.unwrap();
+                            Ok(HttpResponseOk(()))
+                        }
+                        Err(err) => {
+                            ctx.abort_transaction().await.unwrap();
+                            Err(err.into())
+                        }
+                    }
+                }
+            };
+
+            let endpoint_registration = quote! {
+                api.register(checkout_server::payment_webhook).unwrap();
+            };
+
+            (endpoint_definition, endpoint_registration)
+        }
+        None => (quote!(), quote!()),
+    };
+
     let gen = quote! {
         mod checkout_server {
             pub use std::sync::Arc;
@@ -90,6 +171,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
             pub use checkout_core::db::DBClient;
 
             use std::env;
+            use std::ops::Deref;
             use tokio::task;
             use tokio::time::timeout;
             use checkout_core::cloud_pubsub;
@@ -890,6 +972,8 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                 }
             }
+
+            #payment_webhook_endpoint_def
         }
 
         #[async_trait]
@@ -915,6 +999,8 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 api.register(checkout_server::callback_payment).unwrap();
                 api.register(checkout_server::cancel_payment).unwrap();
 
+                #payment_webhook_endpoint_reg
+
                 api
             }
 
